/*
 The software in this package is distributed under the GNU General
 Public License version 2 (with a special exception described below).

 A copy of GNU General Public License (GPL) is included in this distribution,
 in the file COPYING.GPL.

 As a special exception, if other files instantiate templates or use macros
 or inline functions from this file, or you compile this file and link it
 with other works to produce a work based on this file, this file
 does not by itself cause the resulting work to be covered
 by the GNU General Public License.

 However the source code for this file must still be made available
 in accordance with section (3) of the GNU General Public License.

 This exception does not invalidate any other reasons why a work based
 on this file might be covered by the GNU General Public License.
 */

#include "icmp/icmppacketfactory.h"

// for dumping packets
#include <stdlib.h>
#include <ctime>
#include <iostream>
#include <sstream>
#include <boost/scoped_array.hpp>

#include <logfunc.hpp>

#include "boost_assert_handler.h"
#include "icmp/icmpchecksum.h"
#include "icmp/icmpdata.h"
#include "icmp/icmptype.h"
#include "icmp/icmpv4header.h"
#include "icmp/icmpv4packet.h"
#include "icmp/icmpv6header.h"
#include "icmp/icmpv6packet.h"

using namespace std;
using boost::asio::ip::icmp;
using I2n::Logger::GlobalLogger;

//-----------------------------------------------------------------------------
// Dump of data packets to pcap files
//-----------------------------------------------------------------------------
#pragma pack(push, 1) // exact fit -- no padding of data in structs

// pcap header for dumping of packet data
// (http://wiki.wireshark.org/Development/LibpcapFileFormat)
typedef struct pcap_hdr_s {
        uint32_t magic_number;   /* magic number */
        uint16_t version_major;  /* major version number */
        uint16_t version_minor;  /* minor version number */
        int32_t  thiszone;       /* GMT to local correction */
        uint32_t sigfigs;        /* accuracy of timestamps */
        uint32_t snaplen;        /* max length of captured packets, in octets */
        uint32_t network;        /* data link type */
} pcap_hdr_t;

typedef struct pcaprec_hdr_s {
        uint32_t ts_sec;         /* timestamp seconds */
        uint32_t ts_usec;        /* timestamp microseconds */
        uint32_t incl_len;       /* number of octets of packet saved in file */
        uint32_t orig_len;       /* actual length of packet */
} pcaprec_hdr_t;


void dump_packet(const std::string &data)
{
    std::size_t data_size = data.size();

    // create unique file name
    std::stringstream temp_name;
    temp_name << "/datastore/pingcheck.broken/icmp_";
    time_t capture_time = time(0);
    temp_name << capture_time;
    temp_name << "_XXXXXX.pcap";
    std::string temp_str = temp_name.str();
    std::size_t temp_size = temp_str.size();
    boost::scoped_array<char> secure_filename( new char[temp_size + 1] );
    std::copy(temp_str.begin(), temp_str.end(), secure_filename.get());
    secure_filename[temp_size] = '\0';
    int fd = mkstemps(secure_filename.get(), 5);   // 5 = ".pcap".length
    if (fd == -1)
    {
        GlobalLogger.warning() << "Failed to create temp file!" << endl;
        return;
    }

    // create a pcap header
    pcap_hdr_t pcap_hdr;
    pcap_hdr.magic_number = 0xa1b2c3d4;
    pcap_hdr.version_major = 2;
    pcap_hdr.version_minor = 4;
    pcap_hdr.thiszone = 0;
    pcap_hdr.sigfigs = 0;
    pcap_hdr.snaplen = 40;
    pcap_hdr.network = 101;   // LINKTYPE_RAW, according to
                              // http://www.tcpdump.org/linktypes.html

    // create a record (packet) header
    pcaprec_hdr_t pcaprec_hdr;
    pcaprec_hdr.ts_sec = static_cast<uint32_t>(capture_time);
    pcaprec_hdr.ts_usec = 0;    // do not care about that precision
    pcaprec_hdr.incl_len = data_size;
    pcaprec_hdr.orig_len = data_size;

    // create file pointer for file descriptor
    FILE *fp = fdopen(fd, "w");

    // write data
    fwrite(&pcap_hdr, sizeof(pcap_hdr_t), 1, fp);
    fwrite(&pcaprec_hdr, sizeof(pcaprec_hdr_t), 1, fp);
    fwrite(data.c_str(), sizeof(char), data_size, fp);

    // done -- clean up
    fclose(fp);
    close(fd);
    GlobalLogger.debug() << "Dumped a copy of the data into "
                         << secure_filename << endl;
}
#pragma pack(pop) // restore old value

//-----------------------------------------------------------------------------
// IcmpPacketFactory
//-----------------------------------------------------------------------------

bool dump_broken_packets = true;
bool dump_all_packets = false;

/**
 * @brief Creates an ICMP packet from the input stream @c std::istream.
 *
 * @param protocol The packet's network layer protocol, IPv4 or IPv6.
 * @param is The input stream.
 *
 * @return An ICMP Packet object.
 */
IcmpPacketItem IcmpPacketFactory::create_icmp_packet(
        const icmp::socket::protocol_type &protocol,
        istream &is
)
{
    IcmpPacketItem icmp_packet;

    if ( icmp::v4() == protocol )
    {
        icmp_packet.reset( new Icmpv4Packet() );
    }
    else if ( icmp::v6() == protocol )
    {
        icmp_packet.reset( new Icmpv6Packet() );
    }
    else
    {
        GlobalLogger.warning() << "ICMP packet creation failed: "
            << "Unknown protocol, expect ICMP v4 or v6!" << endl;
        icmp_packet.reset();    // --> (!icmp_packet) is true
        return icmp_packet;
    }

    IcmpPacket::ReadReturnCode return_code;

    // create buffer for saving data in it
    stringbuf data_backup;

    // read packet from stream, possibly copying data first
    if (dump_broken_packets | dump_all_packets)
    {
        // read all data into backup
        ostream backup_filler(&data_backup);
        backup_filler << is.rdbuf();
        //GlobalLogger.debug() << "Copied ICMP raw data ("
        //     << data_backup.str().size();
        //     << " bytes) for dumping in case of broken packet" << endl;

        // create a new stream from backup buffer
        //   and use that in read
        istream is_2(&data_backup);
        return_code = icmp_packet->read( is_2 );
    }
    else
        return_code = icmp_packet->read( is );

    if ( return_code != IcmpPacket::ReadReturnCode_OK )
    {
        GlobalLogger.warning() << "ICMP packet creation failed: "
            << IcmpPacket::return_code_to_string(return_code) << endl;
        icmp_packet.reset();    // --> (!icmp_packet) is true
           // --> icmp_pinger will not try to continue working with this packet
    }
    else if ( !is.good() )
    {
        GlobalLogger.warning() << "ICMP packet creation failed: "
            << "Stream not good at end of creation!" << endl;
        icmp_packet.reset();    // --> (!icmp_packet) is true
    }

    // dump data if had trouble with packet
    if ( (dump_broken_packets && !icmp_packet) | dump_all_packets )
        dump_packet(data_backup.str());

    return icmp_packet;
}

/**
 * @brief Creates an ICMP Echo Request packet.
 *
 * @param protocol The packet's network layer protocol, IPv4 or IPv6.
 * @param identifier The packet's identifier number to aid in matching Echo
 * Replies to this Echo Request. May be zero.
 * @param sequence_number The packet's sequence number to aid in matching Echo
 * Replies to this Echo Request. May be zero.
 *
 * @return An ICMP Echo Request packet object.
 */
IcmpPacketItem IcmpPacketFactory::create_icmp_packet_echo_request(
        const icmp::socket::protocol_type &protocol,
        const uint16_t identifier,
        const uint16_t sequence_number
)
{
    BOOST_ASSERT( (icmp::v4() == protocol) || (icmp::v6() == protocol) );

    IcmpPacketItem icmp_packet;

    if ( icmp::v4() == protocol )
    {
        icmp_packet = create_icmpv4_packet_echo_request( identifier, sequence_number );
    }
    else if ( icmp::v6() == protocol )
    {
        icmp_packet = create_icmpv6_packet_echo_request( identifier, sequence_number );
    }
    else
    {
        BOOST_ASSERT( !"Invalid ICMP Packet Type." );                                                                                               //lint !e506
    }

    return icmp_packet;
}

/**
 * @brief Creates an ICMPv4 Echo Request packet.
 *
 * @param identifier The packet's identifier number.
 * @param sequence_number The packet's sequence number.
 *
 * @return An ICMPv4 Echo Request packet object.
 */
IcmpPacketItem IcmpPacketFactory::create_icmpv4_packet_echo_request(
        const uint16_t identifier,
        const uint16_t sequence_number
)
{
    const IcmpData icmp_data( "ping-message" );

    Icmpv4Type type = Icmpv4Type_EchoRequest;
    uint8_t code = 0;
    IcmpChecksum calculator( icmp_data.begin(), icmp_data.end() );
    uint16_t checksum = calculator.compute(
            type, code, identifier, sequence_number
    );
    Icmpv4Header icmp_header(
            type, code, checksum, identifier, sequence_number
    );
    IcmpPacketItem icmp_packet( new Icmpv4Packet( icmp_header, icmp_data ) );

    return icmp_packet;
}

/**
 * @brief Creates an ICMPv6 Echo Request packet.
 *
 * @param identifier The packet's identifier number.
 * @param sequence_number The packet's sequence number.
 *
 * @return An ICMPv6 Echo Request packet object.
 */
IcmpPacketItem IcmpPacketFactory::create_icmpv6_packet_echo_request(
        const uint16_t identifier,
        const uint16_t sequence_number
)
{
    const IcmpData icmp_data( "ping-message" );

    Icmpv6Type type = Icmpv6Type_EchoRequest;
    uint8_t code = 0;
    IcmpChecksum calculator( icmp_data.begin(), icmp_data.end() );
    uint16_t checksum = calculator.compute(
            type, code, identifier, sequence_number
    );
    Icmpv6Header icmp_header(
            type, code, checksum, identifier, sequence_number
    );
    IcmpPacketItem icmp_packet( new Icmpv6Packet( icmp_header, icmp_data ) );

    return icmp_packet;
}


